// ======================================================================
//
// camera.cpp
// jeff grills
//
// copyright 1996 origin systems inc
//
// the prefix "oo" in front of a variable is short for "one over."  So,
// you can expect ooNear to equal 1.0f/near.
//
// the field of view is the whole angle, not half.
//
// ======================================================================

#include "First.h"
#include "Camera.h"

#include "Appear.h"
#include "ExitChain.h"
#include "Extents.h"
#include "Gl.h"
#include "ObjectList.h"

// =====================================================================
// build options

#define USE_DWORD_VIEWPORT       1

// ======================================================================
// constants

const real DEFAULT_NEAR             = CONST_REAL(1);
const real DEFAULT_FAR              = CONST_REAL(5000);
const real DEFAULT_FOV              = Pi_1_3;

// ======================================================================
// make profiling of object list merging easier

#ifndef NPROFILE
const char * const mergeText[4] = 
{
	"merge objlists 1",
	"merge objlists 2",
	"merge objlists 3",
	"merge objlists 4+",
};
#endif

// ======================================================================

Camera::RenderStats   Camera::renderStatsObject;
Camera::RenderStats  *Camera::renderStats = &Camera::renderStatsObject;

// list of lights that can affect the scene
int                   Camera::renderLightMax;
int                   Camera::renderLightCount;
Light               **Camera::renderLight;

// list of objects in the scene that require rendering
bool                  Camera::renderObjectLocked(true);
int                   Camera::renderObjectCount;
int                   Camera::renderObjectMax;
Camera::RenderObject *Camera::renderObject;

// ======================================================================

void Camera::install(void)
{
	renderObjectMax = USE_CAMERA_OBJECTS_INITIAL;
	renderObject    = NEW RenderObject[renderObjectMax];

	renderLightMax  = USE_CAMERA_LIGHTS_INITIAL;
	renderLight     =  NEW Light *[renderLightMax];

	ExitChain::add(&Camera::remove, "camera", 0);
}

// ----------------------------------------------------------------------

void Camera::remove(void)
{
	delete [] renderObject;
	renderObject = NULL;

	delete [] renderLight;
	renderLight = NULL;
}

// ======================================================================

Camera::Camera(int newMaxObjectLists)
: Object(ObjIdCamera)
{
	DEBUG_FATAL(newMaxObjectLists < 0, ("Camera:: negative number of object lists"));

	// set the fov to zero so we don't attempt a setup() until everything is done
	fieldOfView = CONST_REAL(0);

	// setup the camera rendering attributes	
	excludeObject = NULL;

	// allocate the object lists list
	numObjectLists = 0;
	maxObjectLists = newMaxObjectLists;
	objectList     = NEW const ObjectList *[maxObjectLists];

	// projection-independent stuff
	setViewport(0, 0, Gl::width()-1, Gl::height()-1);
	setNear(DEFAULT_NEAR);
	setFar(DEFAULT_FAR);

	// do the fov last because we don't want setup() run until everything is inited
	setFieldOfView(DEFAULT_FOV);
}

// ----------------------------------------------------------------------

Camera::~Camera(void)
{
	numObjectLists = 0;
	maxObjectLists = 0;
	delete [] objectList;
	objectList = NULL;
}

// ----------------------------------------------------------------------
// add an object list to the camera's render list

void Camera::addObjectList(const ObjectList *objList)
{
	int i, j;

	DEBUG_FATAL(numObjectLists == maxObjectLists, ("Cam::addObjectList full"));

	// search for location to add object list
	for (i = 0; i < numObjectLists && objectList[i]->getDepth() < objList->getDepth(); i++)
		;

	// move object lists up to make room for sorted insert
	if (i < numObjectLists)
		for (j = numObjectLists; j > i; j--)
			objectList[j] = objectList[j-1];
	
	objectList[i] = objList;
	numObjectLists++;
}

// ----------------------------------------------------------------------
// remove an object list from the camera's render list

void Camera::removeObjectList(const ObjectList *objList)
{
	int i;

	// search for this object list
	for (i = 0; i < numObjectLists && objectList[i] != objList; i++)
		;

	// if it was found, copy all the elements down a slot
	if (i < numObjectLists)
	{
		for (numObjectLists--; i < numObjectLists; i++)
			objectList[i] = objectList[i+1];
		objectList[numObjectLists] = NULL;
	}
}

// ----------------------------------------------------------------------
// calculate all the camera render values for its current settings

void Camera::setup(void)
{
	real   workingFOV, sinFOV, cosFOV;
	Vector normal;

	// do not do the rest of the calculations until everything has been initialized
	if (fieldOfView == CONST_REAL(0))
		return;

	renderStats->setupCalls++;

	// set the plane d coefficients
	frustumPlaneD[0] = nearPlane;
	frustumPlaneD[1] = -farPlane;
	frustumPlaneD[2] = 0;
	frustumPlaneD[3] = 0;
	frustumPlaneD[4] = 0;
	frustumPlaneD[5] = 0;

	// set the near and far plane normals
	frustumNormal[0].set(CONST_REAL(0), CONST_REAL(0), CONST_REAL(-1));
	frustumNormal[1].set(CONST_REAL(0), CONST_REAL(0), CONST_REAL(1));

	// calculate the left and right plane normals
	workingFOV = fieldOfView * CONST_REAL(0.5);
	sinFOV     = sin(workingFOV);
	cosFOV     = cos(workingFOV);
	frustumNormal[3].set( cosFOV, CONST_REAL(0), -sinFOV);
	frustumNormal[5].set(-cosFOV, CONST_REAL(0), -sinFOV);

	// calculate the top and bottom plane normals
	workingFOV = (workingFOV * halfHeight) / halfWidth;
	sinFOV     = sin(workingFOV);
	cosFOV     = cos(workingFOV);
	frustumNormal[2].set(CONST_REAL(0),  cosFOV, -sinFOV);
	frustumNormal[4].set(CONST_REAL(0), -cosFOV, -sinFOV);

	// xMult = 1/x
	// ax + by + cz + d = 0;
	// x = (by + cz + d) / -a
	// xMult = -a / (by + cz + d)
	clipWidthMultiplier = -frustumNormal[3].x / (frustumNormal[3].y + frustumNormal[3].z + frustumPlaneD[3]);
	clipHeightMultiplier = -frustumNormal[2].y / (frustumNormal[2].x + frustumNormal[2].z + frustumPlaneD[2]);

	viewportWidthMultiplier  = clipWidthMultiplier  * halfWidth;
	viewportHeightMultiplier = clipHeightMultiplier * -halfHeight;
}

// ----------------------------------------------------------------------
// set the near plane distance

void Camera::setNear(real newNearPlane)
{
	DEBUG_FATAL((newNearPlane < 0), ("Cam::setNear %f", VSPRINTF_REAL(newNearPlane)));
	nearPlane = newNearPlane;
	setup();
}

// ----------------------------------------------------------------------
// set the far plane distance

void Camera::setFar(real newFarPlane)
{
	DEBUG_FATAL((newFarPlane < 0), ("Cam::setup far %f", VSPRINTF_REAL(newFarPlane)));
	farPlane = newFarPlane;
	setup();
}

// ----------------------------------------------------------------------
// set the field of view

void Camera::setFieldOfView(real fovRadians)
{
	fieldOfView = fovRadians;
	setup();
}

// ----------------------------------------------------------------------
// set the viewport screen position

void Camera::setViewport(int newX0, int newY0, int newX1, int newY1)
{
	DEBUG_FATAL(newX0 < 0 || newY0 < 0 || newX1 >= Gl::width() || newY1 >= Gl::height(), ("Cam::setViewport off screen"));

#if USE_DWORD_VIEWPORT
	DEBUG_FATAL((newX0 & 3) != 0,  ("Cam::setViewport non-dword x0"));
	DEBUG_FATAL((newX1 & 3) != 3,  ("Cam::setViewport non-dword x1"));
#endif

	x0 = newX0;
	y0 = newY0;
	x1 = newX1;
	y1 = newY1;

	halfWidth    = (x1 + 1 - x0) / 2;
	halfHeight   = (y1 + 1 - y0) / 2;
	centerX      = (x0 + x1 + 1) / 2;  
	centerY      = (y0 + y1 + 1) / 2;

	setup();
}

// ----------------------------------------------------------------------
/// Camera/Functions/Project/bool Camera::project(const Vector &vector, real *x, real *y, Space space, bool offscreenAllowed) const
///
/// Project a vector into screen space
///
/// Return Value:
///
/// Whether the vector was sucessfully projected or not
///
/// Parameters:
///
/// vector The vector to project
///
/// x Pointer to a value in which to store the x coordinate of the projected vertex
///
/// y Pointer to a value in which to store the y coordinate of the projected vertex
///
/// space Indicates the space that the source vector is representated in.  It can be either
/// WorldSpace or CameraSpace.
///
/// offscreenAllowed Determines if the routine will attempt to project the vertex if its
/// result will not be contained within the viewport.
///
/// Remarks:
///
/// Always test the return value of this routine.  Not all source vectors can be projected,
/// because they may be behind the near clipping plane.  Attempting to project such a vector
/// may cause a divide-by-zero error.
///
/// If the source vector is in world space, this routine will need to transform it into
/// camera space.
/// 
/// If the offscreenAllowed parameter is true, then any vector in front of the near plane
/// will be projected.  The resulting screen coordinates may not be within the viewport.
///
/// Example:
///
/// void drawTargetingBox(const Vector &worldPosition, const Camera &camera)
/// {
///		real x, y;
///	
///		if (camera.project(worldPosition, &x, &y, Camera::WorldSpace, false))
///		{
///			Gl::setMaterial(MaterialList::red());
///			Gl::rect(x-5, y-5, x+5, y+5);
///		}
/// }

bool Camera::project(const Vector &vector, real *x, real *y, Space space, bool offscreenAllowed) const
{
	Vector cameraVector;
	real   ooz;

	renderStats->projectCalls++;

	if (space == WorldSpace)
		objectToWorld.inverseRotateTranslate(vector, &cameraVector);
	else
		cameraVector = vector;
	
	// 3d clip the point against the viewing frustum
	if (cameraVector.z < nearPlane || (!offscreenAllowed && (cameraVector.z > farPlane || abs(cameraVector.x)*clipWidthMultiplier > cameraVector.z || abs(cameraVector.y)*clipHeightMultiplier > cameraVector.z)))
	{
		// set unprojected things to somewhere gross so people notice them
		*x = CONST_REAL(320);
		*y = CONST_REAL(240);
		return false;
	}

	ooz = RECIP(cameraVector.z);
	*x  = (cameraVector.x * ooz * viewportWidthMultiplier) + centerX;
	*y  = (cameraVector.y * ooz * viewportHeightMultiplier) + centerY;
	
	return true;
}

// ----------------------------------------------------------------------

bool Camera::project(const Vector &vector, int *x, int *y, Space space, bool offscreenAllowed) const
{
	real xx, yy;

	if (project(vector, &xx, &yy, space, offscreenAllowed))
	{
		*x = REAL_TO_INT(xx);
		*y = REAL_TO_INT(yy);
		return true;
	}

	return false;
}

// ----------------------------------------------------------------------

void Camera::setupGL(void)
{
	renderStats->setupGLCalls++;
	Gl::beginRender(*this);
}

// ----------------------------------------------------------------------
// render all the object lists

bool Camera::render(bool needSetupGL)
{
	int       i, level;
	Vector    p0, p6;

	if (needSetupGL)
		setupGL();

	// render all the object lists, merging all lists with the same depth value
	for (i = 0; i < numObjectLists; )
	{
#ifndef NPROFILE
		const char *merge;
		merge = mergeText[(i < 4) ? i : 3];
		PROFILE_START(merge);
#endif

		// reset stuff for a sorted group of objects
		renderObjectCount = 0;
		renderLightCount  = 0;

		// merge all object lists at this depth
		PROFILE_START("objectList addToCamera");
		renderObjectLocked = false;
		for (level = objectList[i]->getDepth(); i < numObjectLists && objectList[i]->getDepth() == level; ++i)
		{
			// add all the objects in this object list
			objectList[i]->addToCamera(*this);
		}
		renderObjectLocked = true;
		PROFILE_STOP("objectList addToCamera");

		// terminate the linked list of objects
		if (renderObjectCount)
		{
			renderObject[renderObjectCount-1].next = NULL;

			// ray trace the extents
			if (DFlag(DF_E_WIRE_EXTENTS))
				renderExtents(renderObject);

			// setup the lights affecting these polygons
			Gl::setLightList(renderLight, renderLightCount);

			PROFILE_CALL(sortTransformClipAndProjectObjects(renderObject), "sort tcp&p objects");

			// whack the light list
			Gl::setLightList(NULL, 0);
		}

#ifndef NPROFILE
		PROFILE_STOP(merge);
#endif
	}

	Gl::endRender();

	Object::render(false);

	return true;
}

// ----------------------------------------------------------------------

bool Camera::testSphereVisibility(const Vector &worldPoint, real radius, int testPlanes) const
{
	int    i, j;
	Vector cameraPoint;

	// transform the point into camera space
	rotateTranslateW2O(worldPoint, &cameraPoint);

	// test against all sides of the viewing frustum
	for (i = 0, j = 1; i < FRUSTUM_PLANES; i++, j <<= 1) //lint !e701 // left shift of signed value
		if ((testPlanes & j) && frustumNormal[i].dot(cameraPoint) + frustumPlaneD[i] > radius)
		{
			renderStats->failedSphereTests++;
			return false;
		}

	renderStats->passedSphereTests++;
	return true;
}

// ----------------------------------------------------------------------
// add a light to the render

void Camera::addLight(Light *light)
{
	if (renderLightCount >= renderLightMax)
	{
#if USE_CAMERA_LIGHTS_GROW == 0
		DEBUG_FATAL(true, ("Camera::addLight full"));
		return;
#else
		// create the new render light array
		Light **newRenderLight;
		renderLightMax += USE_CAMERA_LIGHTS_GROW;
		newRenderLight = NEW Light *[renderLightMax];

		// copy the light list over
		memcpy(newRenderLight, renderLight, renderLightCount * sizeof(*newRenderLight));

		// replace the old list with the new list
		delete [] renderLight;
		renderLight = newRenderLight;

		WARNING(true, ("Camera::addLight number of render lights grown to %d\n", renderLightMax));
#endif
	}

	renderLight[renderLightCount++] = light;
	renderStats->lightsRendered++;
}

// ----------------------------------------------------------------------
// add an object to the render

void Camera::addObject(const Object *object, const Appear *objectAppear)
{
	RenderObject *ro;

	DEBUG_FATAL(renderObjectLocked, ("Camera::addObject locked"));

	// don't add null objects or excluded objects
	if (!object || object == excludeObject)
	{
		renderStats->objectsCulled++;
		return;
	}

	// check for room in our object list	
	if (renderObjectCount >= renderObjectMax)
	{
#if USE_CAMERA_OBJECTS_GROW == 0
		DEBUG_FATAL(true, ("Camera::addObject list full"));
		renderStats->objectsCulled++;
		return;
#else
		RenderObject *newRenderObject;

		// allocate a new list for rendered objects
		renderObjectMax += USE_CAMERA_OBJECTS_GROW;
		newRenderObject = NEW RenderObject[renderObjectMax];

		// copy the light list over
		memcpy(newRenderObject, renderObject, renderObjectCount * sizeof(*newRenderObject));

		// replace the old list with the new list
		delete [] renderObject;
		renderObject = newRenderObject;

		WARNING(true, ("Camera::addObject number of render objects grown to %d\n", renderObjectMax));
#endif
	}

	// record list of objects to be rendered
	ro                = renderObject+(renderObjectCount++);
	ro->appear        = objectAppear;
	ro->objectToWorld = &object->getObjectToWorldXform();
	ro->next          = ro+1;
	
	renderStats->objectsRendered++;
}
	
// ----------------------------------------------------------------------

void Camera::renderExtents(RenderObject *list)
{
	const int COLORS = 7;

	const Extents        *extents;
	Vector                cameraPosition, result, start, end;
	int                   x, y, region;
	const StaticMaterial *color[COLORS];
	real                  workingFOV, dx, dy, widthMeters, heightMeters, ooHW, ooHH;
	Xform                 xform1, xform2;

	// setup the color table
	color[0] = MaterialList::fetch(  0,   0, 255);
	color[1] = MaterialList::fetch(  0, 255,   0);
	color[2] = MaterialList::fetch(  0, 255, 255);
	color[3] = MaterialList::fetch(255,   0,   0);
	color[4] = MaterialList::fetch(255,   0, 255);
	color[5] = MaterialList::fetch(255, 255,   0);
	color[6] = MaterialList::fetch(255, 255, 255);

	// calculate the width and height of the view window
	workingFOV = fieldOfView * CONST_REAL(0.5);
	widthMeters  = tan(workingFOV) * nearPlane;
	heightMeters = tan(workingFOV * real(halfHeight) / real(halfWidth)) * nearPlane;

	ooHW = CONST_REAL(1) / real(halfWidth);
	ooHH = CONST_REAL(1) / real(halfHeight);

	if (DFlag(DF_E_WIRE_EXTENTS))
		for (; list; list = list->next)
		{
			extents = list->appear->getExtents();

			xform2.invert(*list->objectToWorld);
			xform1.multiply(xform2, objectToWorld);

			if (extents)
			{
				// get the camera position in the object space
				getWorldPosition(&cameraPosition);
				list->objectToWorld->inverseRotateTranslate(cameraPosition, &cameraPosition);

				for (y = -halfHeight; y < halfHeight; ++y)
					for (x = -halfWidth; x < halfWidth; ++x)
					{
						// calculate the camera space position of the pixel on near plane
						dx = real(x) * ooHW;
						dy = real(-y) * ooHH;
						start.set(dx * widthMeters, dy * heightMeters, nearPlane);

						// extent the ray to the far plane
						end.mul(start, farPlane);

						// convert points into target object space
						xform1.rotateTranslate(start, &start);
						xform1.rotateTranslate(end, &end);

						if (extents->collide(start, end, &result, &region))
						{
							Gl::setMaterial(color[region % COLORS]);
							Gl::point(x + centerX, y + centerY);
						}
					}
			}
		}

	for (x = 0; x < COLORS; ++x)
		MaterialList::release(color[x]);
}

// ----------------------------------------------------------------------

void Camera::sortTransformClipAndProjectObjects(RenderObject *list)
{
	RenderObject *sorted, *o, *back, *front;

	PROFILE_START("sort objects");

		// sort the objects based on appearance token
		sorted = NULL;
		while (list)
		{
			// pull an element out of the unsorted list
			o = list;
			list = list->next;

			// search for the proper place to insert it
			for (back = NULL, front = sorted; front && o->appear->getToken() < front->appear->getToken(); back = front, front = front->next)
				;

			// hook it into the sorted list
			o->next = front;
			if (back)
				back->next = o;
			else
				sorted = o;
		}

	PROFILE_STOP("sort objects");

	// render the objects
	for (; sorted; sorted = sorted->next)
		sorted->appear->render(*this, *sorted->objectToWorld);
}

// ----------------------------------------------------------------------

void Camera::RenderStats::reset(void)
{
	memset(this, 0, sizeof(*this));
}

// ----------------------------------------------------------------------

void Camera::RenderStats::debugDump(void)
{
	DPRINT(("cam: %d=setup %d=setupGL %d=project %d=matrix\n", 
		setupCalls, setupGLCalls, projectCalls, matrixMultiplies));
	
	DPRINT(("cam: %d/%d=objs %d/%d=lights %d/%d=sphere %d=vert\n",
		objectsRendered, objectsRendered+objectsCulled,
		lightsRendered,  lightsRendered+lightsCulled,
		passedSphereTests, passedSphereTests+failedSphereTests,
		verticesTransformed));

	DPRINT(("cam: %d/%d=poly %d=back %d=split %d=fill %6.2f=fill\n",
		polygonRenderCount,	polygonTotalCount,
		polygonBackfacingCount,
		polygonSplitCount,
		polygonRenderArea,
		(real(polygonRenderArea) * CONST_REAL(100)) / (real(Gl::width()) * real(Gl::height()))));

	DPRINT(("cam: %d/%d=clip %d=initOut %d=clipOut %d=edges\n",
		polygonClipCount, polygonClipCount+polygonUnclippedCount,
		polygonClipInitialOutCount,
		polygonClipOutCount,
		polygonClipEdgeCount));

	DPRINT(("cam: %d=tranlucent %d=p16 %d=p8 %d=p4 %d=aff\n",
		translucentPolygonsDrawn,
		polygonsPerspective16,
		polygonsPerspective8,
		polygonsPerspective4,
		polygonsAffine));
}

// ----------------------------------------------------------------------

void Camera::debugDump(void)
{
	DPRINT(("cam: %s\n", name));
	renderStats->debugDump();
}

// ----------------------------------------------------------------------
